#!/usr/bin/env python
"""
Compute LHS experiment/sampling design by leveraging the built-in
routines from DAKOTA.

NOTE - be sure to change the hard-coded 'variables' in the input file
       specified in this script!

"""
import logging
logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger()

import pandas as pd
import sys
from subprocess import check_output, STDOUT

from argparse import ArgumentParser, RawDescriptionHelpFormatter

python_script = """
#!/usr/bin/env python

def dakota_interface_list(**kwargs):

    num_fns  = kwargs['functions']

    retval = dict([])
    func_eval = [0., ]*num_fns

    if num_fns == 1:
        retval['fns'] = [func_eval, ]
    else:
        retval['fns'] = [val for val in func_eval]

    return retval
"""

dakota_input = """
environment
    tabular_graphics_data
        tabular_graphics_file 'lhs.dat'

method
    id_method 'lhs sample'
    sampling
        sample_type lhs
    samples = {n}

variables

    uniform_uncertain 13
        lower_bounds -3.0 -5.0 -5.0 -5.0 -5.0 -5.0 -3.0 -3.0 -3.0 0.1 -2.0 240.0 50000.0
        upper_bounds 4.0 4.0 4.0 2.0 1.0 1.0 0.0 -1.0 -1.0 0.6 1.0 310.0 105000.0
        descriptors 'logN_ACC' 'logN_MOS' 'logN_MBS' 'logN_DST01' 'logN_DST02' 'logN_SSLT01' 'logmu_ACC' 'logmu_MOS' 'logmu_MBS' 'kappa_MOS' 'logV' 'T' 'P'

interface
    python
        analysis_drivers = 'lhs_run:dakota_interface_list'

responses
    descriptors 'Smax' 'Neq' 'Nkn'
    response_functions 3
    no_gradients
    no_hessians
"""

parser = ArgumentParser(description=__doc__,
                        formatter_class=RawDescriptionHelpFormatter)
parser.add_argument("-o", "--output", type=str, default="design.csv",
                    help="Name of output file containing the design DataFrame")
parser.add_argument("-n", "--n", type=int, default=100,
                    help="Number of design samples to compute")

if __name__ == "__main__":

    args = parser.parse_args()

    # Write the dummy linked analysis script
    logger.info("Writing lhs_run.py")
    with open("lhs_run.py", "wb") as f:
        f.write(python_script)

    # Write the DAKOTA input file
    logger.info("Writing lhs_in")
    with open("lhs.in", "wb") as f:
        f.write(dakota_input.format(n=args.n))

    # Execute DAKOTA
    logger.info("Computing {} LHS samples via DAKOTA".format(args.n))
    check_output(['dakota', '-i',  'lhs.in'], stderr=STDOUT)

    # Clean output
    logger.info("Reading output from lhs.dat")
    df = pd.read_table('lhs.dat', delim_whitespace=True)
    cols = df.columns[1:-3] # Get rid of eval_id and response cols
    df = df[cols]

    # Save to disk
    logger.info("Saving to {}".format(args.output))
    df.to_csv(args.output)
