#!/usr/bin/env python
"""
Quickly plot comparisons between activation parameterizations and reference
parcel model output for designs processed with the chaos expansion toolkit.

"""
from __future__ import print_function

from functools import partial
import os, pickle
import numpy as np
import pandas as pd

import sklearn.metrics as skm
from statsmodels.distributions import ECDF

import matplotlib.pyplot as plt
import seaborn as sns

#: Plot types
PLOT_TYPES = ["oneone", "dists", ]
#: Variables
PLOT_VARS = ["Nderiv", "Smax"]
#: Save dir
PLOT_DIR = "figs/"

from argparse import ArgumentParser, RawDescriptionHelpFormatter
parser = ArgumentParser(description=__doc__,
                        formatter_class=RawDescriptionHelpFormatter)
parser.add_argument("plot", type=str, choices=PLOT_TYPES,
                    help="type of plot to create")
parser.add_argument("exp_name", type=str,
                    help="name of PCM experiment to reference")

parser.add_argument("-r", "--reference", type=str, required=True,
                    help="Reference sample design to plot")
parser.add_argument("-s", "--scheme", type=str, required=True,
                    help="Scheme to plot; if number, will try to find"
                         " matching chaos expansion")
parser.add_argument("-v", "--var", type=str, required=True,
                    choices=PLOT_VARS,
                    help="Activation variable to plot")

parser.add_argument("--lims", nargs=2, type=float,
                    default=[1e-5, 1e-1],
                    help="x- and y-axis lims (default [1e-5, 1e-1])")

save_group = parser.add_mutually_exclusive_group()
save_group.add_argument("-i", "--interact", action="store_true",
                        help="Enable interactive mode (draw plots to screen)")
save_group.add_argument("--save", action='store_false',
                        help="Save output figure")


def compute_stats(obs, act):
    mae = skm.mean_absolute_error(act, obs)
    r2  = skm.r2_score(act, obs)
    rmse  = np.sqrt(np.sum((obs-act)**2.)/len(act))
    nrmse = rmse/np.sqrt(np.sum((act**2.)/len(act)))

    rel_err = 100.*(obs - act)/act
    ## Mask egregiously high values (1000% error) which screw up the spread
    rel_err = rel_err[np.abs(rel_err) <= 1000.]
    mre = np.mean(rel_err)
    mre_std = np.std(rel_err)

    stats = {
        'mae': mae, 'r2': r2, 'rmse': rmse, 'nrmse': nrmse,
        'mre': mre, 'mre_std': mre_std,
    }

    return stats


def plot_dists_base(param, parcel, var_name, param_name, lims,
                    loglog=True, exp_name = '', **kwargs):
    pct_levs = np.linspace(0., 100., 11)

    fig, axs = plt.subplots(1, 2, figsize=(10, 4))
    plt.subplots_adjust(wspace=0.25, bottom=0.15)
    ax_cdf, ax_pdf = axs

    ## Compute empirical CDFs
    param_cdf = ECDF(param.ravel())
    param_percentiles = [np.percentile(param.ravel(), x) for x in pct_levs]
    parcel_cdf = ECDF(parcel.ravel())
    parcel_percentiles = [np.percentile(parcel.ravel(), x) for x in pct_levs]

    ax_cdf.plot(parcel_percentiles, pct_levs/100., color='k', lw=5,
                   label="parcel model")
    ax_cdf.plot(param_percentiles, pct_levs/100., "--o", ms=8, label=param_name)

    ax_cdf.legend(loc='best')
    ax_cdf.set_xlim(*lims)
    ax_cdf.set_xlabel(var_name)
    if loglog:
        ax_cdf.set_xscale('log')
    ax_cdf.set_ylim(0, 1)
    ax_cdf.set_ylabel("Cumulative Probability")

    ## PDFs
    plot_kwargs = {}
    if loglog:
        plot_kwargs['hist'] = True
        plot_kwargs['kde'] = False
        plot_kwargs['bins'] = np.logspace(*[np.log10(l) for l in lims])
    else:
        plot_kwargs['hist'] = False
        plot_kwargs['kde'] = True
        plot_kwargs['bins'] = np.linspace(*lims)
    ax_pdf = sns.distplot(parcel,
                          color='k', label="parcel model", ax=ax_pdf,
                          kde_kws={'lw': 5},
                          **plot_kwargs)
    ax_pdf = sns.distplot(param,
                          label=param_name, ax=ax_pdf,
                          kde_kws={'linestyle': 'dashed'},
                          **plot_kwargs)
    ax_pdf.set_xlim(*lims)
    ax_pdf.set_xlabel(var_name)
    if loglog:
        ax_pdf.set_xscale('log')
    ax_pdf.set_ylabel("Probability Density")
    ax_pdf.legend(loc="best")

    return [fig, ax_cdf, ax_pdf, ]

def plot_oneone_base(param, parcel, var_name, param_name, lims, coloring='k',
                    loglog=True, ax=None, error_pct=0.5, exp_name='',
                    **kwargs):
    if not ax:
        fig = plt.figure(figsize=(10, 6))
        ax = fig.add_subplot(111)

    # Mask for infinites
    mask = np.isfinite(parcel)
    parcel = parcel[mask]
    param  = param[mask]
    if isinstance(coloring, (pd.Series, np.ndarray, list)):
        coloring = coloring[mask]

    ax.scatter(parcel, param, marker='.', s=30, c=coloring,
               edgecolor='none', alpha=0.8, cmap=plt.get_cmap("OrRd"), **kwargs)
    oo = np.linspace(lims[0], lims[1], 100)
    ax.plot(oo, oo, color='grey', lw=3)
    ax.plot(oo, oo*error_pct, color='k', lw=1, alpha=0.8)
    ax.plot(oo, oo*(1.+error_pct), color='k', lw=1, alpha=0.8)

    ax.set_xlim(lims)
    ax.set_ylim(lims)
    if loglog:
        ax.loglog()

    ax.set_xlabel("%s, parcel model" % var_name)
    ax.set_ylabel("%s, %s" % (var_name, param_name))

    stats = compute_stats(param, parcel)
    ax.text(0.05, 0.775, stat_label.format(**stats),
            transform=ax.transAxes, fontsize=12)

    return [fig, ax]

###########################################################
## Plot aesthetics and configuration

sns.set(style="darkgrid", context="talk", palette="Dark2")
# plt.rc('text', usetex=True)
# plt.rc('font', family='serif')

def is_int(s):
    try:
        _ = int(s)
    except ValueError:
        return False
    return True

def fn_out_fix(s, chars="_${}"):
    for char in chars:
        s = s.replace(char, "")
    return s

fn_fix = lambda s: s.replace("_", "\_")
def param_name(s):
    if s in ["ARG", "MBN"]:
        return s
    else:
        bits = s.split("_")
        order = int(bits[-1])
        return "PCE order %d" % order

stat_label = "RMSE: {rmse:1.2f} ({nrmse:1.2f})\n" + \
             " MAE: {mae:1.2f}\n"  + \
             " R$^2$: {r2:1.2f}\n" + \
             " MRE: {mre:2.2f}$\%$ ({mre_std:2.2f}$\%$)"

z_func = lambda z: 10.**z

###########################################################

if __name__ == "__main__":

    args = parser.parse_args()
    print("Creating %s plot for %s" % (args.plot, args.exp_name))
    print("   Variable:", args.var)
    print("   Sample data:", args.reference)
    if args.interact:
        print("   Interactive mode")
        plt.ion()
    else:
        plt.ioff()
    print("\n")

    # Overwrite plotting function
    plot_kwargs = dict(exp_name=args.exp_name)
    if args.plot == 'oneone':
        plot_kwargs['loglog'] = True
    plot_func = partial(globals()['plot_%s_base' % args.plot],
                        **plot_kwargs)

    # Unload the configured experiment
    exp_dict = pickle.load(open("%s_exp.dict" % args.exp_name, 'r'))
    results_dict = pickle.load(open("%s_results.dict" % args.exp_name, 'r'))
    design_df = pd.read_csv("%s.csv" % args.reference, index_col=0)
    results_df = pd.read_csv("%s_results.csv" % args.reference, index_col=0)

    # Set up args for the plotting routine
    var_key = var_name = args.var
    lims = args.lims

    if var_key == 'Nderiv':
        parcel_data = z_func(results_df['Neq_parcel'])
    else:
        parcel_data = z_func(results_df['%s_parcel' % var_key])

    if is_int(args.scheme):
        scheme_name = "expansion_order_%s" % args.scheme
    else:
        scheme_name = args.scheme
    scheme_data = z_func(results_df['%s_%s' % (var_key, scheme_name)])

    # Mask egregious outlier data
    data_df = pd.DataFrame({'scheme': scheme_data,
                            'parcel': parcel_data})
    # Note that the data is all log data, so we're masking truly outlier
    # points here
    data_df[data_df > 1e10] = np.nan
    data_df[data_df < 0.] = np.nan
    data_df.dropna(inplace=True)

    # Call plotting routine
    fig_ax = plot_func(data_df['scheme'], data_df['parcel'],
                   var_name, param_name(scheme_name), lims)

    fig = fig_ax[0]

    if args.interact:
        plt.show()
    else:
        fn = "{plot}_{var}_{scheme}_{reference}.png".format(
            plot=args.plot, var=var_name, scheme=scheme_name,
            reference=args.reference
        )
        out_path = os.path.join(PLOT_DIR, fn)
        print("Saving to %s" % out_path)
        plt.savefig(out_path, transparent=True, dpi=200,
                    bbox_inches='tight')
